# 49 "src/kernel_internals/parsing/clexer.mll"
 
open Cparser
module H = Hashtbl
module E = Errorloc

let currentLoc () = E.currentLoc ()

let one_line_ghost = ref false
let is_oneline_ghost () = !one_line_ghost
let enter_oneline_ghost () = one_line_ghost := true
let exit_oneline_ghost () = one_line_ghost := false

let ghost_code = ref false
let is_ghost_code () = !ghost_code
let enter_ghost_code () = ghost_code := true
let exit_ghost_code () = ghost_code := false

let ghost_annot = ref false
let is_ghost_annot () = !ghost_annot
let enter_ghost_annot () = ghost_annot := true
let exit_ghost_annot () = ghost_annot := false

let addComment c = Cabshelper.Comments.add (currentLoc()) c

(* Some debugging support for line numbers *)
let dbgToken (t: token) =
  if false then begin
    let dprintf fmt = Kernel.debug fmt in
    (match t with
         IDENT n -> dprintf "IDENT(%s)\n" n
       | LBRACE l -> dprintf "LBRACE(%d)\n" (fst l).Filepath.pos_lnum
       | RBRACE l -> dprintf "RBRACE(%d)\n" (fst l).Filepath.pos_lnum
       | IF l -> dprintf "IF(%d)\n" (fst l).Filepath.pos_lnum
       | SWITCH l -> dprintf "SWITCH(%d)\n" (fst l).Filepath.pos_lnum
       | RETURN l -> dprintf "RETURN(%d)\n" (fst l).Filepath.pos_lnum
       | _ -> ()) ;
    t
  end else
    t

(*
** Keyword hashtable
*)
let lexicon = H.create 211
let init_lexicon _ =
  H.clear lexicon;
  Logic_env.reset_typenames ();
  Logic_env.builtin_types_as_typenames ();
  List.iter
    (fun (key, builder) -> H.add lexicon key builder)
    [ ("auto", fun loc -> AUTO loc);
      ("const", fun loc -> CONST loc);
      ("__const", fun loc -> CONST loc);
      ("__const__", fun loc -> CONST loc);
      ("static", fun loc -> STATIC loc);
      ("extern", fun loc -> EXTERN loc);
      ("long", fun loc -> LONG loc);
      ("short", fun loc -> SHORT loc);
      ("register", fun loc -> REGISTER loc);
      ("signed", fun loc -> SIGNED loc);
      ("__signed", fun loc -> SIGNED loc);
      ("unsigned", fun loc -> UNSIGNED loc);
      ("volatile", fun loc -> VOLATILE loc);
      ("__volatile", fun loc -> VOLATILE loc);
      (* WW: see /usr/include/sys/cdefs.h for why __signed and __volatile
       * are accepted GCC-isms *)
      ("char", fun loc -> CHAR loc);
      ("_Bool", fun loc -> BOOL loc);
      ("int", fun loc -> INT loc);
      ("float", fun loc -> FLOAT loc);
      ("double", fun loc -> DOUBLE loc);
      ("void", fun loc -> VOID loc);
      ("enum", fun loc -> ENUM loc);
      ("struct", fun loc -> STRUCT loc);
      ("typedef", fun loc -> TYPEDEF loc);
      ("union", fun loc -> UNION loc);
      ("break", fun loc -> BREAK loc);
      ("continue", fun loc -> CONTINUE loc);
      ("goto", fun loc -> GOTO loc);
      ("return", fun loc -> dbgToken (RETURN loc));
      ("switch", fun loc -> dbgToken (SWITCH loc));
      ("case", fun loc -> CASE loc);
      ("default", fun loc -> DEFAULT loc);
      ("while", fun loc -> WHILE loc);
      ("do", fun loc -> DO loc);
      ("for", fun loc -> FOR loc);
      ("if", fun loc -> dbgToken (IF loc));
      ("else", fun _ -> ELSE);
      (*** Implementation specific keywords ***)
      ("__signed__", fun loc -> SIGNED loc);
      ("__inline__", fun loc -> INLINE loc);
      ("inline", fun loc -> INLINE loc);
      ("__inline", fun loc -> INLINE loc);
      ("_inline",
       fun loc ->
         if !Cprint.msvcMode then
           INLINE loc
         else begin
           Kernel.(
             warning
               ~wkey:wkey_conditional_feature
               "_inline is a MSVC keyword, \
                use a msvc-specific machdep to enable it");
           IDENT ("_inline")
         end);
      ("_Noreturn",
       fun loc ->
         if Kernel.C11.get () then NORETURN loc
         else begin
           Kernel.(
             warning
               ~wkey:wkey_conditional_feature
               "_Noreturn is a C11 keyword, use -c11 option to enable it");
           IDENT "_Noreturn"
         end);
      ("_Static_assert",
       fun loc ->
         if Kernel.C11.get () then STATIC_ASSERT loc
         else begin
           Kernel.(
             warning
               ~wkey:wkey_conditional_feature
               "_Static_assert is a C11 keyword, use -c11 option to enable it");
           IDENT "_Static_assert"
         end);
      ("__attribute__", fun loc -> ATTRIBUTE loc);
      ("__attribute", fun loc -> ATTRIBUTE loc);
      ("__blockattribute__", fun _ -> BLOCKATTRIBUTE);
      ("__blockattribute", fun _ -> BLOCKATTRIBUTE);
      ("__asm__", fun loc -> ASM loc);
      ("asm", fun loc -> ASM loc);
      ("__typeof__", fun loc -> TYPEOF loc);
      ("__typeof", fun loc -> TYPEOF loc);
      ("typeof", fun loc -> TYPEOF loc);
      ("__alignof", fun loc -> ALIGNOF loc);
      ("__alignof__", fun loc -> ALIGNOF loc);
      ("__volatile__", fun loc -> VOLATILE loc);
      ("__volatile", fun loc -> VOLATILE loc);

      ("__FUNCTION__", fun loc -> FUNCTION__ loc);
      ("__func__", fun loc -> FUNCTION__ loc); (* ISO 6.4.2.2 *)
      ("__PRETTY_FUNCTION__", fun loc -> PRETTY_FUNCTION__ loc);
      ("__label__", fun _ -> LABEL__);
      (*** weimer: GCC arcana ***)
      ("__restrict", fun loc -> RESTRICT loc);
      ("restrict", fun loc -> RESTRICT loc);
(*      ("__extension__", EXTENSION); *)
      (**** MS VC ***)
      ("__int64", fun _ -> INT64 (currentLoc ()));
      ("__int32", fun loc -> INT loc);
      ("_cdecl",  fun _ -> MSATTR ("_cdecl", currentLoc ()));
      ("__cdecl", fun _ -> MSATTR ("__cdecl", currentLoc ()));
      ("_stdcall", fun _ -> MSATTR ("_stdcall", currentLoc ()));
      ("__stdcall", fun _ -> MSATTR ("__stdcall", currentLoc ()));
      ("_fastcall", fun _ -> MSATTR ("_fastcall", currentLoc ()));
      ("__fastcall", fun _ -> MSATTR ("__fastcall", currentLoc ()));
      ("__w64", fun _ -> MSATTR("__w64", currentLoc ()));
      ("__declspec", fun loc -> DECLSPEC loc);
      ("__forceinline", fun loc -> INLINE loc); (* !! we turn forceinline
                                                 * into inline *)
      ("__try", fun loc -> TRY loc);
      ("__except", fun loc -> EXCEPT loc);
      ("__finally", fun loc -> FINALLY loc);
      (* weimer: some files produced by 'GCC -E' expect this type to be
       * defined *)
      ("__builtin_va_list",
       fun _ -> NAMED_TYPE "__builtin_va_list");
      ("__builtin_va_arg", fun loc -> BUILTIN_VA_ARG loc);
      ("__builtin_types_compatible_p", fun loc -> BUILTIN_TYPES_COMPAT loc);
      ("__builtin_offsetof", fun loc -> BUILTIN_OFFSETOF loc);
      ("_Thread_local",
       fun loc ->
         if Kernel.C11.get () then THREAD_LOCAL loc
         else begin
           Kernel.(
             warning
               ~wkey:wkey_conditional_feature
               "_Thread_local is a C11 keyword, use -c11 option to enable it");
           IDENT "_Thread_local"
         end);
      (* We recognize __thread for GCC machdeps *)
      ("__thread",
       fun loc ->
         if Cil.gccMode () then
           THREAD loc
         else begin
           Kernel.(
             warning
               ~wkey:wkey_conditional_feature
               "__thread is a GCC extension, use a GCC-based machdep to enable it");
           IDENT "__thread"
         end);
      ("__FC_FILENAME__",
       (fun loc ->
          let filename =
            Filepath.Normalized.to_pretty_string (fst loc).pos_path
          in
          let seq = String.to_seq filename in
          let convert_char c = Int64.of_int (Char.code c) in
          let l =
            Seq.fold_left (fun acc c -> convert_char c :: acc) [] seq
          in
          CST_STRING (List.rev l,loc)));
      (* The following C11 tokens are not yet supported, so we provide some
         helpful error messages. Usage of 'fatal' instead of 'error' below
         prevents duplicate error messages due to parsing errors. *)
      ("_Alignas",
       fun loc ->
         Kernel.fatal ~source:(fst loc)
           "_Alignas is currently unsupported by Frama-C.");
      ("_Alignof",
       fun loc ->
         Kernel.fatal ~source:(fst loc)
           "_Alignof is currently unsupported by Frama-C.");
      ("_Complex",
       fun loc ->
         Kernel.fatal ~source:(fst loc)
           "_Complex is currently unsupported by Frama-C.");
      ("_Generic",
       fun loc ->
         Kernel.fatal ~source:(fst loc)
           "_Generic is currently unsupported by Frama-C.");
      ("_Imaginary",
       fun loc ->
         Kernel.fatal ~source:(fst loc)
           "_Imaginary is currently unsupported by Frama-C.");
     ]


let is_c_keyword s = Hashtbl.mem lexicon s

(* Mark an identifier as a type name. The old mapping is preserved and will
 * be reinstated when we exit this context *)
let add_type name =
   (* ignore (print_string ("adding type name " ^ name ^ "\n"));  *)
  H.add lexicon name (fun _ -> NAMED_TYPE name);
  Logic_env.add_typename name

let context : string list list ref = ref [ [] ]

let push_context _ = context := []::!context

let pop_context _ =
  match !context with
    [] -> Kernel.fatal "Empty context stack"
  | con::sub ->
		(context := sub;
		List.iter (fun name ->
                            (* Format.eprintf
                              "removing lexicon for %s@." name; *)
                             H.remove lexicon name;
                             Logic_env.remove_typename name
                          ) con)

(* Mark an identifier as a variable name. The old mapping is preserved and
 * will be reinstated when we exit this context  *)
let add_identifier name =
  match !context with
    [] -> Kernel.fatal "Empty context stack"
  | con::sub ->
      (context := (name::con)::sub;
       (*Format.eprintf "adding IDENT for %s@." name;*)
       H.add lexicon name (fun _ -> dbgToken (IDENT name));
       Logic_env.hide_typename name
      )


(*
** Useful primitives
*)
let scan_ident id =
  let here = currentLoc () in
  try (H.find lexicon id) here
  (* default to variable name, as opposed to type *)
  with Not_found -> dbgToken (IDENT id)


(*
** Buffer processor
*)


let init ~(filename: string) : Lexing.lexbuf =
  init_lexicon ();
  (* Initialize the pointer in Errormsg *)
  Lexerhack.add_type := add_type;
  Lexerhack.push_context := push_context;
  Lexerhack.pop_context := pop_context;
  Lexerhack.add_identifier := add_identifier;
  E.startParsing filename


let finish () =
  E.finishParsing ();
  Logic_env.reset_typenames ()


(*** escape character management ***)
let scan_escape (char: char) : int64 =
  let result = match char with
    'n' -> '\n'
  | 'r' -> '\r'
  | 't' -> '\t'
  | 'b' -> '\b'
  | 'f' -> '\012'  (* ASCII code 12 *)
  | 'v' -> '\011'  (* ASCII code 11 *)
  | 'a' -> '\007'  (* ASCII code 7 *)
  | 'e' | 'E' -> '\027'  (* ASCII code 27. This is a GCC extension *)
  | '\'' -> '\''
  | '"'-> '"'     (* '"' *)
  | '?' -> '?'
  | '(' when not !Cprint.msvcMode -> '('
  | '{' when not !Cprint.msvcMode -> '{'
  | '[' when not !Cprint.msvcMode -> '['
  | '%' when not !Cprint.msvcMode -> '%'
  | '\\' -> '\\'
  | other ->
    E.parse_error "Unrecognized escape sequence: \\%c" other
  in
  Int64.of_int (Char.code result)

let scan_hex_escape str =
  let radix = Int64.of_int 16 in
  let the_value = ref Int64.zero in
  (* start at character 2 to skip the \x *)
  for i = 2 to (String.length str) - 1 do
    let thisDigit = Cabshelper.valueOfDigit (String.get str i) in
    (* the_value := !the_value * 16 + thisDigit *)
    the_value := Int64.add (Int64.mul !the_value radix) thisDigit
  done;
  !the_value

let scan_oct_escape str =
  let radix = Int64.of_int 8 in
  let the_value = ref Int64.zero in
  (* start at character 1 to skip the \x *)
  for i = 1 to (String.length str) - 1 do
    let thisDigit = Cabshelper.valueOfDigit (String.get str i) in
    (* the_value := !the_value * 8 + thisDigit *)
    the_value := Int64.add (Int64.mul !the_value radix) thisDigit
  done;
  !the_value

let lex_hex_escape remainder lexbuf =
  let prefix = scan_hex_escape (Lexing.lexeme lexbuf) in
  prefix :: remainder lexbuf

let lex_oct_escape remainder lexbuf =
  let prefix = scan_oct_escape (Lexing.lexeme lexbuf) in
  prefix :: remainder lexbuf

let lex_simple_escape remainder lexbuf =
  let lexchar = Lexing.lexeme_char lexbuf 1 in
  let prefix = scan_escape lexchar in
  prefix :: remainder lexbuf

let lex_unescaped remainder lexbuf =
  let prefix = Int64.of_int (Char.code (Lexing.lexeme_char lexbuf 0)) in
  prefix :: remainder lexbuf

let lex_comment remainder buffer lexbuf =
  let s = Lexing.lexeme lexbuf in
  if s = "\n" then E.newline() ;
  (match buffer with None -> () | Some b -> Buffer.add_string b s) ;
  remainder buffer lexbuf

let do_ghost_else_comments register comments =
  let ends_with_n =
    let last = String.length comments in
    (* note that comments contains at least a blank *)
    '\n' = String.get comments (last - 1)
  in
  let comments = String.split_on_char '\n' comments in
  let comments = List.map String.trim comments in
  let process_line s =
    E.newline () ;
    if register
    && Kernel.PrintComments.get ()
    && not (String.equal "" s)
    then
      addComment s
  in
  let rec iter f = function
    | [] -> ()
    (* We ignore the last line if it does not contain "\n" as it means that
      it directly precedes the "else" and thus contains only whitespaces *)
    | [ _ ] when not ends_with_n -> ()
    | x :: l -> f x ; iter f l
  in
  iter process_line comments

let do_lex_comment ?(first_string="") remainder lexbuf =
  let buffer =
    if Kernel.PrintComments.get () then begin
      let b = Buffer.create 80 in
      Buffer.add_string b first_string;
      Some b
    end else None
  in remainder buffer lexbuf ;
  match buffer with
    | Some b -> addComment (Buffer.contents b)
    | None -> ()


(* ISO standard locale-specific function to convert a wide character
 * into a sequence of normal characters. Here we work on strings.
 * We convert L"Hi" to "H\000i\000"
  matth: this seems unused.
let wbtowc wstr =
  let len = String.length wstr in
  let dest = String.make (len * 2) '\000' in
  for i = 0 to len-1 do
    dest.[i*2] <- wstr.[i] ;
  done ;
  dest
*)

(* This function converts the "Hi" in L"Hi" to { L'H', L'i', L'\0' }
  matth: this seems unused.
let wstr_to_warray wstr =
  let len = String.length wstr in
  let res = ref "{ " in
  for i = 0 to len-1 do
    res := !res ^ (Printf.sprintf "L'%c', " wstr.[i])
  done ;
  res := !res ^ "}" ;
  !res
*)

(* Pragmas get explicit end-of-line tokens.
 * Elsewhere they are silently discarded as whitespace. *)
let pragmaLine = ref false

let annot_char = ref '@'

let () =
  Kernel.ReadAnnot.add_set_hook
    (fun _ x ->
      (* prevent the C lexer interpretation of comments *)
      annot_char := if x then '@' else '\000')

let annot_start_pos = ref Cabshelper.cabslu
let buf = Buffer.create 1024

let save_current_pos () =
  annot_start_pos := currentLoc ()

let annot_lex initial rule lexbuf =
  try
    save_current_pos ();
    Buffer.clear buf;
    rule lexbuf
  with Parsing.Parse_error ->
    let source = Cil_datatype.Position.of_lexing_pos (Lexing.lexeme_start_p lexbuf) in
    Kernel.warning ~wkey:Kernel.wkey_annot_error ~source "skipping annotation";
    initial lexbuf

let make_annot ~one_line default lexbuf s =
  let start = snd !annot_start_pos in
  match Logic_lexer.annot (start, s) with
  | Some (stop, token) ->
    lexbuf.Lexing.lex_curr_p <- Cil_datatype.Position.to_lexing_pos stop;
    (* The filename has already been normalized, so we must reuse it "as is". *)
    E.setCurrentFile (stop.Filepath.pos_path :> string);
    E.setCurrentLine stop.Filepath.pos_lnum;
    if one_line then E.newline ();
    (match token with
     | Logic_ptree.Adecl d -> DECL d
     | Logic_ptree.Aspec -> SPEC (start,s)
     (* At this point, we only have identified a function spec. Complete
           parsing of the annotation will only occur in the cparser.mly rule.
     *)
     | Logic_ptree.Acode_annot (loc,a) -> CODE_ANNOT (a, loc)
     | Logic_ptree.Aloop_annot (loc,a) -> LOOP_ANNOT (a,loc)
     | Logic_ptree.Aattribute_annot (loc,a) -> ATTRIBUTE_ANNOT (a, loc))
  | None -> (* error occured and annotation is discarded. Find a normal token. *)
    default lexbuf


# 482 "src/kernel_internals/parsing/clexer.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base =
   "\000\000\177\255\178\255\091\000\192\000\185\255\186\255\187\255\
    \188\255\191\255\192\255\195\255\196\255\197\255\002\000\199\255\
    \035\000\067\000\093\000\070\000\094\000\091\000\072\000\169\000\
    \173\000\074\000\013\001\025\001\063\001\242\255\136\001\244\255\
    \223\001\076\000\247\255\249\255\074\001\213\000\093\001\208\255\
    \230\255\173\000\092\001\254\255\059\000\255\255\252\255\094\000\
    \253\255\248\255\001\000\032\000\038\000\035\000\069\000\251\255\
    \193\255\229\255\246\255\054\002\141\002\228\002\059\003\146\003\
    \233\003\064\004\151\004\238\004\069\005\156\005\243\005\074\006\
    \161\006\248\006\079\007\166\007\253\007\084\008\171\008\002\009\
    \089\009\176\009\007\010\094\010\181\010\012\011\241\255\243\255\
    \172\000\089\011\117\011\127\011\164\000\052\001\062\001\139\011\
    \161\011\168\000\156\011\158\011\173\011\183\000\077\001\076\001\
    \058\001\172\000\098\001\153\011\175\011\199\011\173\000\240\255\
    \004\012\014\012\043\012\053\012\212\001\082\012\130\002\092\012\
    \217\002\153\012\191\012\173\000\178\000\008\001\163\011\121\012\
    \005\013\229\012\038\013\048\003\071\013\135\003\238\255\168\000\
    \156\000\125\000\183\000\078\012\182\000\122\001\020\001\165\011\
    \239\255\007\001\162\000\131\000\198\000\205\000\234\255\220\255\
    \138\000\183\000\212\000\003\000\235\255\214\255\233\255\212\255\
    \213\255\232\255\231\255\203\255\228\255\204\255\227\255\226\255\
    \190\255\194\255\219\255\024\001\225\255\218\255\027\001\224\255\
    \221\255\189\255\115\013\202\013\033\014\120\014\207\014\042\001\
    \255\255\094\001\253\255\254\255\043\001\255\255\019\002\253\255\
    \254\255\044\001\255\255\065\015\249\255\250\255\053\001\252\255\
    \253\255\086\002\255\255\251\255\070\015\249\255\239\000\250\000\
    \080\015\009\004\255\255\053\001\124\001\065\001\175\011\253\255\
    \064\001\067\001\047\001\106\001\055\001\068\001\252\255\077\001\
    \073\001\086\001\099\001\096\004\129\015\086\001\120\001\203\001\
    \180\001\167\001\168\001\172\001\185\001\187\001\247\001\218\001\
    \219\001\233\001\242\001\248\001\246\001\254\001\251\255\006\002\
    \047\002\014\002\006\002\001\002\046\002\030\002\042\002\037\002\
    \040\002\072\002\094\002\089\002\091\002\093\002\102\002\109\002\
    \090\002\131\002\134\002\118\002\130\002\126\002\176\002\187\002\
    \187\002\187\002\185\002\181\002\191\002\224\002\219\002\201\002\
    \206\002\214\002\009\003\008\003\017\003\036\003\025\003\024\003\
    \014\003\018\003\047\003\041\003\041\003\055\003\058\003\085\003\
    \109\003\091\003\105\003\104\003\103\003\094\003\094\003\136\003\
    \124\003\140\003\143\003\126\003\165\003\160\003\162\003\124\013\
    \251\255\136\015\183\004\255\255\252\255\137\015\155\015\157\015\
    \164\015\253\255\005\013\076\015\080\015\026\004\253\255\254\255\
    \255\255\027\004\254\255\255\255\151\010\250\255\251\255\177\012\
    \255\255\252\255\058\005\209\015\247\015\145\005\253\255\147\010\
    \250\255\251\255\166\015\255\255\252\255\232\005\046\016\119\016\
    \063\006\253\255\037\013\253\255\254\255\193\006\155\003\254\255\
    \255\255\011\011\253\255\178\003\255\255\178\003\161\003\170\003\
    \254\255\094\004\252\255\253\255\190\003\184\003\181\003\208\003\
    \109\016\217\003\023\004\168\016\005\000\007\011\148\010\185\011\
    \196\016\196\011\064\012\065\012\067\012\038\013\206\012\068\012\
    \212\003\228\003\255\255\114\004\252\255\253\255\254\255\054\004\
    \255\255\055\004\255\255\181\004\253\255\002\004\255\003\252\003\
    \007\004\024\007\006\000\111\007\017\004\012\004\056\004\255\255\
    \201\004\253\255\254\255\255\255";
  Lexing.lex_backtrk =
   "\255\255\255\255\255\255\076\000\076\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\057\000\255\255\
    \038\000\040\000\039\000\055\000\053\000\054\000\046\000\045\000\
    \044\000\050\000\071\000\019\000\019\000\255\255\076\000\255\255\
    \076\000\049\000\255\255\255\255\005\000\078\000\048\000\255\255\
    \255\255\002\000\000\000\255\255\001\000\255\255\255\255\003\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\076\000\076\000\076\000\076\000\076\000\
    \076\000\010\000\076\000\076\000\076\000\076\000\073\000\076\000\
    \076\000\076\000\076\000\074\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\075\000\255\255\255\255\
    \255\255\255\255\255\255\015\000\255\255\019\000\019\000\019\000\
    \018\000\255\255\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\255\255\019\000\019\000\019\000\019\000\255\255\255\255\
    \255\255\015\000\255\255\015\000\255\255\015\000\255\255\015\000\
    \255\255\255\255\017\000\255\255\017\000\017\000\017\000\255\255\
    \255\255\255\255\015\000\255\255\015\000\255\255\255\255\017\000\
    \017\000\017\000\255\255\016\000\255\255\016\000\016\000\016\000\
    \255\255\016\000\016\000\016\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\032\000\255\255\255\255\033\000\255\255\
    \255\255\255\255\076\000\076\000\076\000\076\000\072\000\001\000\
    \255\255\255\255\255\255\255\255\002\000\255\255\255\255\255\255\
    \255\255\002\000\255\255\255\255\255\255\255\255\006\000\255\255\
    \255\255\001\000\255\255\255\255\255\255\255\255\006\000\006\000\
    \002\000\001\000\255\255\255\255\002\000\002\000\002\000\255\255\
    \002\000\002\000\002\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\005\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\004\000\001\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\003\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\005\000\
    \255\255\255\255\002\000\003\000\001\000\002\000\255\255\255\255\
    \255\255\255\255\005\000\255\255\255\255\002\000\003\000\001\000\
    \002\000\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\002\000\255\255\255\255\255\255\255\255\
    \255\255\004\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \001\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\003\000\
    \255\255\001\000\255\255\003\000\255\255\255\255\255\255\255\255\
    \255\255\001\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255";
  Lexing.lex_default =
   "\001\000\000\000\000\000\255\255\255\255\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\255\255\000\000\
    \255\255\255\255\000\000\000\000\255\255\255\255\039\000\000\000\
    \000\000\046\000\043\000\000\000\255\255\000\000\000\000\255\255\
    \000\000\000\000\255\255\255\255\255\255\255\255\255\255\000\000\
    \000\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\000\000\000\000\
    \255\255\255\255\255\255\255\255\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\000\000\000\000\255\255\000\000\
    \000\000\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\186\000\000\000\000\000\255\255\000\000\191\000\000\000\
    \000\000\255\255\000\000\196\000\000\000\000\000\255\255\000\000\
    \000\000\255\255\000\000\000\000\205\000\000\000\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\056\001\
    \000\000\061\001\255\255\000\000\000\000\061\001\061\001\061\001\
    \061\001\000\000\067\001\067\001\067\001\070\001\000\000\000\000\
    \000\000\074\001\000\000\000\000\077\001\000\000\000\000\081\001\
    \000\000\000\000\255\255\255\255\255\255\255\255\000\000\088\001\
    \000\000\000\000\092\001\000\000\000\000\255\255\255\255\255\255\
    \255\255\000\000\099\001\000\000\000\000\255\255\103\001\000\000\
    \000\000\106\001\000\000\255\255\000\000\255\255\255\255\255\255\
    \000\000\255\255\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\126\001\126\001\126\001\
    \126\001\126\001\126\001\126\001\126\001\126\001\126\001\126\001\
    \255\255\255\255\000\000\140\001\000\000\000\000\000\000\255\255\
    \000\000\255\255\000\000\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \161\001\000\000\000\000\000\000";
  Lexing.lex_trans =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\036\000\035\000\049\000\036\000\036\000\050\000\123\001\
    \155\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\036\000\000\000\000\000\000\000\000\000\000\000\
    \036\000\025\000\029\000\034\000\156\000\033\000\020\000\031\000\
    \008\000\007\000\022\000\024\000\005\000\023\000\026\000\038\000\
    \028\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\014\000\006\000\018\000\016\000\017\000\015\000\
    \177\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\030\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\010\000\037\000\009\000\019\000\032\000\
    \176\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\004\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\012\000\021\000\011\000\013\000\003\000\
    \173\000\174\000\169\000\167\000\165\000\162\000\058\000\151\000\
    \052\000\057\000\056\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\053\000\054\000\168\000\
    \164\000\171\000\170\000\166\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\045\000\255\255\
    \045\000\055\000\003\000\152\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\160\000\163\000\
    \157\000\048\000\110\000\048\000\139\000\139\000\105\000\049\000\
    \101\000\106\000\050\000\138\000\003\000\091\000\161\000\159\000\
    \134\000\134\000\158\000\134\000\148\000\047\000\144\000\144\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\144\000\150\000\090\000\134\000\136\000\153\000\
    \002\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\123\000\090\000\134\000\137\000\003\000\
    \154\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\178\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\149\000\051\000\113\000\113\000\113\000\
    \113\000\113\000\113\000\113\000\113\000\113\000\113\000\091\000\
    \155\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\036\000\135\000\172\000\036\000\036\000\
    \175\000\184\000\189\000\194\000\144\000\134\000\090\000\203\000\
    \145\000\223\000\091\000\220\000\036\000\094\000\255\255\255\255\
    \091\000\144\000\036\000\219\000\091\000\091\000\093\000\096\000\
    \096\000\096\000\096\000\096\000\096\000\096\000\096\000\027\000\
    \027\000\090\000\091\000\091\000\144\000\134\000\090\000\090\000\
    \108\000\088\000\092\000\090\000\090\000\095\000\255\255\042\000\
    \188\000\144\000\107\000\094\000\041\000\216\000\093\000\215\000\
    \091\000\090\000\090\000\106\000\093\000\215\000\215\000\089\000\
    \101\000\090\000\040\000\215\000\044\000\092\000\215\000\090\000\
    \109\000\088\000\101\000\090\000\090\000\221\000\101\000\090\000\
    \092\000\222\000\086\000\095\000\003\000\255\255\224\000\087\000\
    \225\000\090\000\090\000\106\000\093\000\215\000\215\000\089\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\101\000\226\000\227\000\049\001\146\000\090\000\
    \217\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\140\000\036\001\211\000\147\000\003\000\
    \218\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\115\000\115\000\115\000\115\000\
    \115\000\115\000\115\000\115\000\115\000\115\000\029\001\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\015\001\006\001\005\001\001\001\192\000\252\000\249\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\060\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\248\000\241\000\247\000\193\000\059\000\240\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\242\000\255\255\255\255\187\000\201\000\
    \243\000\244\000\201\000\201\000\245\000\246\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \201\000\246\000\246\000\250\000\251\000\246\000\201\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\253\000\254\000\255\000\000\001\003\000\246\000\068\000\
    \003\000\003\000\003\000\066\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\067\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\117\000\117\000\117\000\117\000\117\000\117\000\
    \117\000\117\000\117\000\117\000\002\001\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\001\
    \004\001\246\000\246\000\007\001\008\001\009\001\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \010\001\011\001\012\001\013\001\003\000\014\001\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\061\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\119\000\119\000\119\000\119\000\119\000\119\000\119\000\
    \119\000\119\000\119\000\192\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\246\000\016\001\
    \017\001\018\001\019\001\020\001\021\001\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\022\001\
    \023\001\024\001\025\001\003\000\026\001\062\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \130\000\130\000\130\000\130\000\130\000\130\000\130\000\130\000\
    \130\000\130\000\027\001\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\028\001\246\000\030\001\
    \031\001\032\001\033\001\034\001\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\035\001\246\000\
    \037\001\038\001\003\000\039\001\003\000\003\000\003\000\003\000\
    \003\000\003\000\063\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\132\000\
    \132\000\132\000\132\000\132\000\132\000\132\000\132\000\132\000\
    \132\000\040\001\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\041\001\042\001\043\001\044\001\
    \045\001\046\001\047\001\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\048\001\246\000\050\001\
    \051\001\003\000\052\001\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\064\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\053\001\054\001\
    \246\000\109\001\209\000\110\001\111\001\209\000\209\000\112\001\
    \104\001\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\209\000\072\001\075\001\117\001\118\001\
    \119\001\209\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\120\001\136\001\125\001\137\001\
    \003\000\138\001\065\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\144\001\146\001\115\001\
    \114\001\228\000\150\001\115\001\228\000\228\000\151\001\152\001\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\228\000\153\001\141\001\157\001\115\001\158\001\
    \228\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\255\255\143\001\159\001\115\001\003\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\148\001\000\000\
    \058\001\000\000\148\001\058\001\058\001\116\001\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\058\001\000\000\163\001\000\000\148\001\000\000\058\001\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\148\001\003\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\076\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\071\001\255\255\149\001\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\003\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \071\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\085\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\085\001\142\001\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\003\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \069\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\086\001\086\001\086\001\086\001\086\001\086\001\086\001\
    \086\001\162\001\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\003\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\070\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \096\001\096\001\096\001\096\001\096\001\096\001\096\001\096\001\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\003\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\097\001\
    \097\001\097\001\097\001\097\001\097\001\097\001\097\001\000\000\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\003\000\000\000\072\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\101\001\000\000\000\000\101\001\101\001\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\101\001\000\000\000\000\000\000\000\000\
    \000\000\101\001\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \003\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \073\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\155\001\000\000\000\000\155\001\155\001\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\155\001\000\000\000\000\000\000\000\000\000\000\
    \155\001\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\003\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\074\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\154\001\000\000\000\000\000\000\
    \155\001\000\000\000\000\155\001\155\001\000\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\155\001\000\000\000\000\000\000\000\000\000\000\155\001\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\003\000\000\000\
    \075\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\154\001\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\156\001\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\003\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\003\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\077\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\003\000\000\000\003\000\003\000\003\000\
    \003\000\078\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\003\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\079\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\003\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\080\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \003\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\081\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\003\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\082\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\003\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\083\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\123\001\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\080\001\091\001\000\000\000\000\084\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\090\001\
    \127\001\000\000\000\000\079\001\000\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\123\001\000\000\000\000\085\000\108\001\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\108\001\255\255\
    \000\000\000\000\000\000\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\107\001\003\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\121\000\
    \108\001\122\000\122\000\122\000\122\000\122\000\122\000\122\000\
    \122\000\122\000\122\000\089\001\255\255\000\000\000\000\078\001\
    \000\000\000\000\122\000\122\000\122\000\122\000\122\000\122\000\
    \120\000\000\000\120\000\000\000\000\000\119\000\119\000\119\000\
    \119\000\119\000\119\000\119\000\119\000\119\000\119\000\113\000\
    \113\000\113\000\113\000\113\000\113\000\113\000\113\000\113\000\
    \113\000\091\000\122\000\122\000\122\000\122\000\122\000\122\000\
    \000\000\000\000\000\000\128\001\112\000\111\000\000\000\091\000\
    \000\000\000\000\091\000\111\000\091\000\000\000\123\001\091\000\
    \090\000\096\000\096\000\096\000\096\000\096\000\096\000\096\000\
    \096\000\027\000\027\000\091\000\000\000\091\000\090\000\000\000\
    \106\000\090\000\000\000\090\000\112\000\111\000\090\000\000\000\
    \103\000\000\000\102\000\111\000\000\000\099\000\106\000\000\000\
    \090\000\000\000\090\000\101\000\090\000\091\000\098\000\107\000\
    \134\000\000\000\144\000\106\000\000\000\000\000\090\000\000\000\
    \106\000\090\000\101\000\090\000\215\000\097\000\090\000\255\255\
    \104\000\000\000\097\000\255\255\090\000\100\000\106\000\135\000\
    \000\000\145\000\090\000\101\000\090\000\127\001\098\000\000\000\
    \134\000\102\000\144\000\216\000\000\000\000\000\000\000\000\000\
    \127\001\000\000\101\000\000\000\215\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\090\000\000\000\000\000\118\000\
    \133\001\118\000\000\000\106\000\117\000\117\000\117\000\117\000\
    \117\000\117\000\117\000\117\000\117\000\117\000\113\000\113\000\
    \113\000\113\000\113\000\113\000\113\000\113\000\113\000\113\000\
    \000\000\000\000\123\001\128\001\000\000\123\001\123\001\000\000\
    \000\000\000\000\000\000\114\000\111\000\000\000\116\000\000\000\
    \116\000\000\000\111\000\115\000\115\000\115\000\115\000\115\000\
    \115\000\115\000\115\000\115\000\115\000\115\000\115\000\115\000\
    \115\000\115\000\115\000\115\000\115\000\115\000\115\000\132\001\
    \000\000\000\000\000\000\114\000\111\000\000\000\000\000\000\000\
    \000\000\000\000\111\000\111\000\000\000\000\000\139\000\139\000\
    \000\000\111\000\117\000\117\000\117\000\117\000\117\000\117\000\
    \117\000\117\000\117\000\117\000\119\000\119\000\119\000\119\000\
    \119\000\119\000\119\000\119\000\119\000\119\000\000\000\000\000\
    \111\000\000\000\142\000\111\000\127\001\127\001\111\000\127\001\
    \127\001\111\000\111\000\141\000\133\000\000\000\133\000\000\000\
    \111\000\132\000\132\000\132\000\132\000\132\000\132\000\132\000\
    \132\000\132\000\132\000\000\000\000\000\000\000\000\000\140\000\
    \111\000\255\255\143\000\000\000\000\000\000\000\111\000\000\000\
    \000\000\000\000\111\000\141\000\255\255\000\000\000\000\000\000\
    \111\000\128\000\128\000\128\000\128\000\128\000\128\000\128\000\
    \128\000\128\000\128\000\000\000\000\000\000\000\000\000\000\000\
    \123\001\000\000\128\000\128\000\128\000\128\000\128\000\128\000\
    \000\000\082\001\082\001\082\001\082\001\082\001\082\001\082\001\
    \082\001\000\000\000\000\000\000\000\000\128\000\000\000\122\000\
    \122\000\122\000\122\000\122\000\122\000\122\000\122\000\122\000\
    \122\000\000\000\128\000\128\000\128\000\128\000\128\000\128\000\
    \122\000\122\000\122\000\122\000\122\000\122\000\000\000\000\000\
    \000\000\083\001\000\000\125\000\000\000\000\000\255\255\127\000\
    \131\000\255\255\131\000\000\000\124\000\130\000\130\000\130\000\
    \130\000\130\000\130\000\130\000\130\000\130\000\130\000\000\000\
    \122\000\122\000\122\000\122\000\122\000\122\000\000\000\060\001\
    \123\000\083\001\127\001\126\000\000\000\000\000\101\001\127\000\
    \123\001\101\001\101\001\135\001\124\000\128\000\128\000\128\000\
    \128\000\128\000\128\000\128\000\128\000\128\000\128\000\101\001\
    \255\255\255\255\000\000\255\255\255\255\101\001\128\000\128\000\
    \128\000\128\000\128\000\128\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\129\000\130\000\130\000\
    \130\000\130\000\130\000\130\000\130\000\130\000\130\000\130\000\
    \000\000\068\001\000\000\000\000\000\000\000\000\128\000\128\000\
    \128\000\128\000\128\000\128\000\111\000\000\000\000\000\000\000\
    \000\000\000\000\111\000\000\000\000\000\129\000\000\000\132\000\
    \132\000\132\000\132\000\132\000\132\000\132\000\132\000\132\000\
    \132\000\000\000\127\001\000\000\000\000\058\001\059\001\000\000\
    \058\001\058\001\000\000\000\000\111\000\111\000\000\000\000\000\
    \000\000\000\000\111\000\111\000\000\000\000\000\058\001\003\000\
    \000\000\134\001\000\000\000\000\058\001\000\000\057\001\000\000\
    \100\001\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\111\000\000\000\000\000\
    \000\000\255\255\000\000\111\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\255\255\000\000\
    \000\000\000\000\003\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\179\000\003\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\255\255\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\255\255\255\255\000\000\
    \000\000\003\000\000\000\003\000\003\000\003\000\003\000\180\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\255\255\000\000\000\000\000\000\
    \003\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \181\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\003\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\182\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\003\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\201\000\202\000\000\000\201\000\201\000\209\000\
    \210\000\000\000\209\000\209\000\000\000\255\255\000\000\000\000\
    \255\255\255\255\000\000\201\000\255\255\000\000\000\000\000\000\
    \209\000\201\000\000\000\197\000\000\000\000\000\209\000\000\000\
    \000\000\200\000\199\000\000\000\000\000\000\000\060\001\000\000\
    \198\000\000\000\066\001\000\000\000\000\000\000\208\000\208\000\
    \208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\
    \208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\
    \208\000\208\000\228\000\000\000\000\000\228\000\228\000\000\000\
    \000\000\255\255\255\255\000\000\255\255\255\255\000\000\000\000\
    \000\000\000\000\000\000\228\000\213\000\000\000\000\000\000\000\
    \000\000\228\000\000\000\000\000\255\255\212\000\255\255\255\255\
    \068\001\255\255\060\001\060\001\068\001\255\255\000\000\000\000\
    \255\255\000\000\207\000\000\000\000\000\000\000\206\000\063\001\
    \063\001\211\000\000\000\000\000\214\000\066\001\000\000\060\001\
    \000\000\000\000\000\000\000\000\000\000\212\000\065\001\000\000\
    \238\000\000\000\063\001\000\000\064\001\000\000\000\000\000\000\
    \000\000\000\000\000\000\064\001\000\000\231\000\093\001\093\001\
    \093\001\093\001\093\001\093\001\093\001\093\001\000\000\000\000\
    \000\000\000\000\234\000\000\000\062\001\062\001\000\000\000\000\
    \230\000\000\000\237\000\000\000\000\000\229\000\000\000\000\000\
    \235\000\000\000\000\000\232\000\236\000\000\000\233\000\062\001\
    \239\000\062\001\000\000\000\000\000\000\000\000\094\001\000\000\
    \062\001\084\001\084\001\084\001\084\001\084\001\084\001\084\001\
    \084\001\084\001\084\001\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\084\001\084\001\084\001\084\001\084\001\084\001\
    \000\000\000\000\000\000\000\000\000\000\000\000\094\001\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\084\001\
    \084\001\084\001\084\001\084\001\084\001\084\001\084\001\084\001\
    \084\001\000\000\084\001\084\001\084\001\084\001\084\001\084\001\
    \084\001\084\001\084\001\084\001\084\001\084\001\000\000\000\000\
    \000\000\255\255\000\000\000\000\000\000\000\000\255\255\000\000\
    \000\000\000\000\000\000\000\000\255\255\000\000\000\000\000\000\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \084\001\084\001\084\001\084\001\084\001\084\001\095\001\095\001\
    \095\001\095\001\095\001\095\001\095\001\095\001\095\001\095\001\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\095\001\
    \095\001\095\001\095\001\095\001\095\001\000\000\123\001\123\001\
    \000\000\123\001\123\001\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\123\001\
    \255\255\255\255\000\000\000\000\000\000\123\001\000\000\095\001\
    \095\001\095\001\095\001\095\001\095\001\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\122\001\255\255\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\000\000\255\255\095\001\
    \095\001\095\001\095\001\095\001\095\001\095\001\095\001\095\001\
    \095\001\123\001\123\001\000\000\123\001\123\001\000\000\000\000\
    \095\001\095\001\095\001\095\001\095\001\095\001\000\000\000\000\
    \000\000\000\000\123\001\000\000\000\000\000\000\000\000\000\000\
    \123\001\124\001\000\000\000\000\000\000\128\001\123\001\000\000\
    \128\001\128\001\121\001\000\000\000\000\000\000\000\000\122\001\
    \095\001\095\001\095\001\095\001\095\001\095\001\128\001\000\000\
    \000\000\000\000\000\000\000\000\128\001\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\130\001\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\124\001\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\121\001\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \131\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\129\001\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255";
  Lexing.lex_check =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\050\000\000\000\000\000\050\000\124\001\
    \154\001\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \000\000\000\000\000\000\000\000\155\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \014\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \016\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \017\000\017\000\018\000\019\000\020\000\022\000\033\000\025\000\
    \051\000\033\000\033\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\052\000\053\000\018\000\
    \021\000\018\000\018\000\020\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\044\000\041\000\
    \044\000\054\000\003\000\025\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\023\000\021\000\
    \024\000\047\000\092\000\047\000\088\000\088\000\097\000\037\000\
    \105\000\110\000\037\000\123\000\004\000\101\000\023\000\023\000\
    \136\000\137\000\024\000\138\000\140\000\041\000\146\000\147\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\148\000\149\000\101\000\135\000\124\000\152\000\
    \000\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\124\000\101\000\135\000\124\000\004\000\
    \153\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\026\000\037\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\027\000\
    \154\000\027\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\036\000\125\000\171\000\036\000\036\000\
    \174\000\183\000\188\000\193\000\145\000\125\000\027\000\198\000\
    \142\000\206\000\093\000\207\000\036\000\027\000\042\000\038\000\
    \104\000\142\000\036\000\211\000\094\000\028\000\027\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\093\000\103\000\102\000\145\000\125\000\027\000\104\000\
    \093\000\028\000\027\000\094\000\028\000\027\000\042\000\038\000\
    \185\000\142\000\094\000\028\000\038\000\213\000\027\000\217\000\
    \106\000\103\000\102\000\094\000\028\000\216\000\213\000\028\000\
    \103\000\093\000\038\000\218\000\042\000\093\000\219\000\104\000\
    \093\000\028\000\102\000\094\000\028\000\220\000\104\000\106\000\
    \028\000\221\000\030\000\028\000\030\000\041\000\223\000\030\000\
    \224\000\103\000\102\000\094\000\028\000\216\000\213\000\028\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\102\000\225\000\226\000\229\000\141\000\106\000\
    \212\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\141\000\230\000\212\000\141\000\030\000\
    \212\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\032\000\116\000\116\000\116\000\116\000\
    \116\000\116\000\116\000\116\000\116\000\116\000\231\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\232\000\233\000\234\000\235\000\190\000\236\000\237\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\238\000\239\000\240\000\190\000\032\000\239\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\059\000\241\000\042\000\038\000\185\000\201\000\
    \242\000\243\000\201\000\201\000\244\000\245\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \201\000\247\000\248\000\249\000\250\000\251\000\201\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\252\000\253\000\254\000\255\000\059\000\000\001\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\060\000\118\000\118\000\118\000\118\000\118\000\118\000\
    \118\000\118\000\118\000\118\000\001\001\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\002\001\
    \003\001\004\001\005\001\006\001\007\001\008\001\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \009\001\010\001\011\001\012\001\060\000\013\001\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \061\000\120\000\120\000\120\000\120\000\120\000\120\000\120\000\
    \120\000\120\000\120\000\190\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\014\001\015\001\
    \016\001\017\001\018\001\019\001\020\001\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\021\001\
    \022\001\023\001\024\001\061\000\025\001\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\062\000\
    \131\000\131\000\131\000\131\000\131\000\131\000\131\000\131\000\
    \131\000\131\000\026\001\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\027\001\028\001\029\001\
    \030\001\031\001\032\001\033\001\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\034\001\035\001\
    \036\001\037\001\062\000\038\001\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\063\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\039\001\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\040\001\041\001\042\001\043\001\
    \044\001\045\001\046\001\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\047\001\048\001\049\001\
    \050\001\063\000\051\001\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\064\000\052\001\053\001\
    \054\001\107\001\209\000\109\001\110\001\209\000\209\000\111\001\
    \102\001\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\209\000\069\001\073\001\116\001\117\001\
    \118\001\209\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\119\001\121\001\122\001\136\001\
    \064\000\137\001\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\065\000\143\001\145\001\113\001\
    \113\001\227\000\149\001\113\001\227\000\227\000\150\001\151\001\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\227\000\152\001\139\001\156\001\113\001\157\001\
    \227\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\102\001\139\001\158\001\113\001\065\000\
    \255\255\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\066\000\255\255\255\255\147\001\255\255\
    \058\001\255\255\147\001\058\001\058\001\113\001\255\255\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\058\001\255\255\160\001\255\255\147\001\255\255\058\001\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\255\255\255\255\255\255\147\001\066\000\255\255\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\067\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\069\001\073\001\147\001\255\255\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\255\255\255\255\255\255\255\255\067\000\255\255\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\068\000\082\001\082\001\082\001\082\001\082\001\082\001\
    \082\001\082\001\139\001\255\255\255\255\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \255\255\255\255\255\255\255\255\068\000\255\255\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \069\000\085\001\085\001\085\001\085\001\085\001\085\001\085\001\
    \085\001\160\001\255\255\255\255\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\255\255\
    \255\255\255\255\255\255\069\000\255\255\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\070\000\
    \093\001\093\001\093\001\093\001\093\001\093\001\093\001\093\001\
    \255\255\255\255\255\255\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\255\255\255\255\
    \255\255\255\255\070\000\255\255\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\071\000\096\001\
    \096\001\096\001\096\001\096\001\096\001\096\001\096\001\255\255\
    \255\255\255\255\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\255\255\255\255\255\255\
    \255\255\071\000\255\255\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\072\000\255\255\255\255\
    \255\255\255\255\101\001\255\255\255\255\101\001\101\001\255\255\
    \255\255\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\101\001\255\255\255\255\255\255\255\255\
    \255\255\101\001\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\255\255\255\255\255\255\255\255\
    \072\000\255\255\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\073\000\255\255\255\255\255\255\
    \255\255\153\001\255\255\255\255\153\001\153\001\255\255\255\255\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\153\001\255\255\255\255\255\255\255\255\255\255\
    \153\001\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\255\255\255\255\255\255\255\255\073\000\
    \255\255\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\074\000\153\001\255\255\255\255\255\255\
    \155\001\255\255\255\255\155\001\155\001\255\255\255\255\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\155\001\255\255\255\255\255\255\255\255\255\255\155\001\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\255\255\255\255\255\255\255\255\074\000\255\255\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\075\000\155\001\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\155\001\255\255\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\255\255\255\255\255\255\255\255\075\000\255\255\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\076\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \255\255\255\255\255\255\255\255\076\000\255\255\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \077\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\255\255\
    \255\255\255\255\255\255\077\000\255\255\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\077\000\078\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\255\255\255\255\
    \255\255\255\255\078\000\255\255\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\078\000\078\000\
    \078\000\078\000\078\000\078\000\078\000\078\000\079\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\255\255\255\255\255\255\
    \255\255\079\000\255\255\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\080\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\080\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\255\255\255\255\255\255\255\255\
    \080\000\255\255\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\081\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\255\255\255\255\255\255\255\255\081\000\
    \255\255\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\082\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\082\000\
    \082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\
    \082\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\
    \082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\
    \082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\
    \082\000\082\000\255\255\255\255\255\255\255\255\082\000\255\255\
    \082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\
    \082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\
    \082\000\082\000\082\000\082\000\082\000\082\000\082\000\082\000\
    \082\000\082\000\083\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\126\001\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\076\001\087\001\255\255\255\255\083\000\255\255\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\083\000\083\000\083\000\083\000\083\000\083\000\083\000\
    \083\000\084\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\087\001\
    \126\001\255\255\255\255\076\001\255\255\255\255\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \255\255\125\001\255\255\255\255\084\000\105\001\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \085\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\105\001\125\001\
    \255\255\255\255\255\255\255\255\255\255\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\255\255\
    \255\255\255\255\105\001\085\000\255\255\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\089\000\
    \105\001\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\089\000\087\001\126\001\255\255\255\255\076\001\
    \255\255\255\255\089\000\089\000\089\000\089\000\089\000\089\000\
    \090\000\255\255\090\000\255\255\255\255\090\000\090\000\090\000\
    \090\000\090\000\090\000\090\000\090\000\090\000\090\000\091\000\
    \091\000\091\000\091\000\091\000\091\000\091\000\091\000\091\000\
    \091\000\095\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \255\255\255\255\255\255\127\001\091\000\091\000\255\255\107\000\
    \255\255\255\255\098\000\091\000\099\000\255\255\129\001\096\000\
    \095\000\096\000\096\000\096\000\096\000\096\000\096\000\096\000\
    \096\000\096\000\096\000\100\000\255\255\108\000\107\000\255\255\
    \095\000\098\000\255\255\099\000\091\000\091\000\096\000\255\255\
    \098\000\255\255\099\000\091\000\255\255\096\000\107\000\255\255\
    \095\000\255\255\100\000\099\000\108\000\109\000\096\000\095\000\
    \126\000\255\255\143\000\108\000\255\255\255\255\107\000\255\255\
    \095\000\098\000\100\000\099\000\214\000\098\000\096\000\125\001\
    \098\000\255\255\096\000\105\001\109\000\096\000\107\000\126\000\
    \255\255\143\000\100\000\099\000\108\000\127\001\096\000\255\255\
    \126\000\100\000\143\000\214\000\255\255\255\255\255\255\255\255\
    \129\001\255\255\100\000\255\255\214\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\109\000\255\255\255\255\112\000\
    \129\001\112\000\255\255\109\000\112\000\112\000\112\000\112\000\
    \112\000\112\000\112\000\112\000\112\000\112\000\113\000\113\000\
    \113\000\113\000\113\000\113\000\113\000\113\000\113\000\113\000\
    \255\255\255\255\130\001\131\001\255\255\132\001\135\001\255\255\
    \255\255\255\255\255\255\113\000\113\000\255\255\114\000\255\255\
    \114\000\255\255\113\000\114\000\114\000\114\000\114\000\114\000\
    \114\000\114\000\114\000\114\000\114\000\115\000\115\000\115\000\
    \115\000\115\000\115\000\115\000\115\000\115\000\115\000\130\001\
    \255\255\255\255\255\255\113\000\113\000\255\255\255\255\255\255\
    \255\255\255\255\113\000\115\000\255\255\255\255\139\000\139\000\
    \255\255\115\000\117\000\117\000\117\000\117\000\117\000\117\000\
    \117\000\117\000\117\000\117\000\119\000\119\000\119\000\119\000\
    \119\000\119\000\119\000\119\000\119\000\119\000\255\255\255\255\
    \117\000\255\255\139\000\115\000\130\001\131\001\117\000\132\001\
    \135\001\115\000\119\000\139\000\127\000\255\255\127\000\255\255\
    \119\000\127\000\127\000\127\000\127\000\127\000\127\000\127\000\
    \127\000\127\000\127\000\255\255\255\255\255\255\255\255\139\000\
    \117\000\127\001\139\000\255\255\255\255\255\255\117\000\255\255\
    \255\255\255\255\119\000\139\000\129\001\255\255\255\255\255\255\
    \119\000\121\000\121\000\121\000\121\000\121\000\121\000\121\000\
    \121\000\121\000\121\000\255\255\255\255\255\255\255\255\255\255\
    \134\001\255\255\121\000\121\000\121\000\121\000\121\000\121\000\
    \255\255\079\001\079\001\079\001\079\001\079\001\079\001\079\001\
    \079\001\255\255\255\255\255\255\255\255\122\000\255\255\122\000\
    \122\000\122\000\122\000\122\000\122\000\122\000\122\000\122\000\
    \122\000\255\255\121\000\121\000\121\000\121\000\121\000\121\000\
    \122\000\122\000\122\000\122\000\122\000\122\000\255\255\255\255\
    \255\255\079\001\255\255\122\000\255\255\255\255\066\001\122\000\
    \129\000\066\001\129\000\255\255\122\000\129\000\129\000\129\000\
    \129\000\129\000\129\000\129\000\129\000\129\000\129\000\255\255\
    \122\000\122\000\122\000\122\000\122\000\122\000\255\255\066\001\
    \122\000\079\001\134\001\122\000\255\255\255\255\098\001\122\000\
    \133\001\098\001\098\001\134\001\122\000\128\000\128\000\128\000\
    \128\000\128\000\128\000\128\000\128\000\128\000\128\000\098\001\
    \130\001\131\001\255\255\132\001\135\001\098\001\128\000\128\000\
    \128\000\128\000\128\000\128\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\128\000\130\000\130\000\
    \130\000\130\000\130\000\130\000\130\000\130\000\130\000\130\000\
    \255\255\066\001\255\255\255\255\255\255\255\255\128\000\128\000\
    \128\000\128\000\128\000\128\000\130\000\255\255\255\255\255\255\
    \255\255\255\255\130\000\255\255\255\255\128\000\255\255\132\000\
    \132\000\132\000\132\000\132\000\132\000\132\000\132\000\132\000\
    \132\000\255\255\133\001\255\255\255\255\055\001\055\001\255\255\
    \055\001\055\001\255\255\255\255\130\000\132\000\255\255\255\255\
    \255\255\255\255\130\000\132\000\255\255\255\255\055\001\178\000\
    \255\255\133\001\255\255\255\255\055\001\255\255\055\001\255\255\
    \098\001\255\255\255\255\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\132\000\255\255\255\255\
    \255\255\079\001\255\255\132\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\134\001\255\255\
    \255\255\255\255\178\000\255\255\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\178\000\178\000\
    \178\000\178\000\178\000\178\000\178\000\178\000\179\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\179\000\179\000\179\000\179\000\179\000\179\000\
    \179\000\179\000\179\000\179\000\255\255\066\001\255\255\255\255\
    \255\255\255\255\255\255\179\000\179\000\179\000\179\000\179\000\
    \179\000\179\000\179\000\179\000\179\000\179\000\179\000\179\000\
    \179\000\179\000\179\000\179\000\179\000\179\000\179\000\179\000\
    \179\000\179\000\179\000\179\000\179\000\098\001\133\001\255\255\
    \255\255\179\000\255\255\179\000\179\000\179\000\179\000\179\000\
    \179\000\179\000\179\000\179\000\179\000\179\000\179\000\179\000\
    \179\000\179\000\179\000\179\000\179\000\179\000\179\000\179\000\
    \179\000\179\000\179\000\179\000\179\000\180\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\180\000\180\000\180\000\180\000\180\000\180\000\180\000\
    \180\000\180\000\180\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\180\000\180\000\180\000\180\000\180\000\180\000\
    \180\000\180\000\180\000\180\000\180\000\180\000\180\000\180\000\
    \180\000\180\000\180\000\180\000\180\000\180\000\180\000\180\000\
    \180\000\180\000\180\000\180\000\055\001\255\255\255\255\255\255\
    \180\000\255\255\180\000\180\000\180\000\180\000\180\000\180\000\
    \180\000\180\000\180\000\180\000\180\000\180\000\180\000\180\000\
    \180\000\180\000\180\000\180\000\180\000\180\000\180\000\180\000\
    \180\000\180\000\180\000\180\000\181\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \181\000\181\000\181\000\181\000\181\000\181\000\181\000\181\000\
    \181\000\181\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\181\000\181\000\181\000\181\000\181\000\181\000\181\000\
    \181\000\181\000\181\000\181\000\181\000\181\000\181\000\181\000\
    \181\000\181\000\181\000\181\000\181\000\181\000\181\000\181\000\
    \181\000\181\000\181\000\255\255\255\255\255\255\255\255\181\000\
    \255\255\181\000\181\000\181\000\181\000\181\000\181\000\181\000\
    \181\000\181\000\181\000\181\000\181\000\181\000\181\000\181\000\
    \181\000\181\000\181\000\181\000\181\000\181\000\181\000\181\000\
    \181\000\181\000\181\000\182\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\182\000\
    \182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\
    \182\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\
    \182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\
    \182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\
    \182\000\182\000\255\255\255\255\255\255\255\255\182\000\255\255\
    \182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\
    \182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\
    \182\000\182\000\182\000\182\000\182\000\182\000\182\000\182\000\
    \182\000\182\000\195\000\195\000\255\255\195\000\195\000\204\000\
    \204\000\255\255\204\000\204\000\255\255\067\001\255\255\255\255\
    \067\001\068\001\255\255\195\000\068\001\255\255\255\255\255\255\
    \204\000\195\000\255\255\195\000\255\255\255\255\204\000\255\255\
    \255\255\195\000\195\000\255\255\255\255\255\255\067\001\255\255\
    \195\000\255\255\068\001\255\255\255\255\255\255\204\000\204\000\
    \204\000\204\000\204\000\204\000\204\000\204\000\204\000\204\000\
    \208\000\208\000\208\000\208\000\208\000\208\000\208\000\208\000\
    \208\000\208\000\228\000\255\255\255\255\228\000\228\000\255\255\
    \255\255\057\001\061\001\255\255\057\001\061\001\255\255\255\255\
    \255\255\255\255\255\255\228\000\208\000\255\255\255\255\255\255\
    \255\255\228\000\255\255\255\255\062\001\208\000\063\001\062\001\
    \067\001\063\001\057\001\061\001\068\001\064\001\255\255\255\255\
    \064\001\255\255\204\000\255\255\255\255\255\255\204\000\057\001\
    \061\001\208\000\255\255\255\255\208\000\062\001\255\255\063\001\
    \255\255\255\255\255\255\255\255\255\255\208\000\064\001\255\255\
    \228\000\255\255\062\001\255\255\063\001\255\255\255\255\255\255\
    \255\255\255\255\255\255\064\001\255\255\228\000\090\001\090\001\
    \090\001\090\001\090\001\090\001\090\001\090\001\255\255\255\255\
    \255\255\255\255\228\000\255\255\057\001\061\001\255\255\255\255\
    \228\000\255\255\228\000\255\255\255\255\228\000\255\255\255\255\
    \228\000\255\255\255\255\228\000\228\000\255\255\228\000\062\001\
    \228\000\063\001\255\255\255\255\255\255\255\255\090\001\255\255\
    \064\001\083\001\083\001\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\083\001\083\001\083\001\083\001\083\001\083\001\
    \255\255\255\255\255\255\255\255\255\255\255\255\090\001\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\084\001\
    \084\001\084\001\084\001\084\001\084\001\084\001\084\001\084\001\
    \084\001\255\255\083\001\083\001\083\001\083\001\083\001\083\001\
    \084\001\084\001\084\001\084\001\084\001\084\001\255\255\255\255\
    \255\255\195\000\255\255\255\255\255\255\255\255\204\000\255\255\
    \255\255\255\255\255\255\255\255\067\001\255\255\255\255\255\255\
    \068\001\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \084\001\084\001\084\001\084\001\084\001\084\001\094\001\094\001\
    \094\001\094\001\094\001\094\001\094\001\094\001\094\001\094\001\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\094\001\
    \094\001\094\001\094\001\094\001\094\001\255\255\120\001\120\001\
    \255\255\120\001\120\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\120\001\
    \057\001\061\001\255\255\255\255\255\255\120\001\255\255\094\001\
    \094\001\094\001\094\001\094\001\094\001\255\255\255\255\255\255\
    \255\255\255\255\255\255\062\001\120\001\063\001\255\255\255\255\
    \255\255\255\255\255\255\255\255\064\001\255\255\090\001\095\001\
    \095\001\095\001\095\001\095\001\095\001\095\001\095\001\095\001\
    \095\001\123\001\123\001\255\255\123\001\123\001\255\255\255\255\
    \095\001\095\001\095\001\095\001\095\001\095\001\255\255\255\255\
    \255\255\255\255\123\001\255\255\255\255\255\255\255\255\255\255\
    \123\001\120\001\255\255\255\255\255\255\128\001\128\001\255\255\
    \128\001\128\001\120\001\255\255\255\255\255\255\255\255\123\001\
    \095\001\095\001\095\001\095\001\095\001\095\001\128\001\255\255\
    \255\255\255\255\255\255\255\255\128\001\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\128\001\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\123\001\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\123\001\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \128\001\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\128\001\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\128\001";
  Lexing.lex_base_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\002\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\004\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000";
  Lexing.lex_backtrk_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000";
  Lexing.lex_default_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000";
  Lexing.lex_trans_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\001\000\000\000\001\000\001\000\001\000\001\000\001\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\001\000\000\000\001\000\000\000\000\000\000\000\
    \001\000\000\000\001\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000";
  Lexing.lex_check_code =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\227\000\255\255\228\000\227\000\227\000\228\000\228\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\227\000\255\255\228\000\255\255\255\255\255\255\
    \227\000\255\255\228\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255";
  Lexing.lex_code =
   "\255\001\255\255\000\001\255";
}

let rec initial lexbuf =
   __ocaml_lex_initial_rec lexbuf 0
and __ocaml_lex_initial_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 587 "src/kernel_internals/parsing/clexer.mll"
                            suf
# 2028 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 2) lexbuf.Lexing.lex_curr_pos in
# 589 "src/kernel_internals/parsing/clexer.mll"
      (
        do_lex_comment ~first_string:suf comment lexbuf ;
        initial lexbuf
      )
# 2035 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
let
# 594 "src/kernel_internals/parsing/clexer.mll"
                        c
# 2041 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 2) in
# 595 "src/kernel_internals/parsing/clexer.mll"
    ( if c = !annot_char then begin
        annot_lex initial annot_first_token lexbuf
      end else
	begin
	  do_lex_comment ~first_string:(String.make 1 c) comment lexbuf ;
          initial lexbuf
	end
    )
# 2052 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
let
# 604 "src/kernel_internals/parsing/clexer.mll"
                            suf
# 2058 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 2) lexbuf.Lexing.lex_curr_pos in
# 605 "src/kernel_internals/parsing/clexer.mll"
    (
      do_lex_comment ~first_string:suf onelinecomment lexbuf ;
      E.newline();
      if is_oneline_ghost () then begin
        exit_oneline_ghost ();
        RGHOST
      end else begin
        initial lexbuf
      end
    )
# 2071 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
let
# 616 "src/kernel_internals/parsing/clexer.mll"
                    c
# 2077 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 2) in
# 617 "src/kernel_internals/parsing/clexer.mll"
    ( if c = !annot_char then begin
        annot_lex initial annot_one_line lexbuf
      end else
	begin
	  do_lex_comment ~first_string:(String.make 1 c) onelinecomment lexbuf;
	  E.newline();
	  if is_oneline_ghost () then
	    begin
              exit_oneline_ghost ();
              RGHOST
	    end
	  else
	    begin
              initial lexbuf
	    end
	end
    )
# 2097 "src/kernel_internals/parsing/clexer.ml"

  | 4 ->
# 635 "src/kernel_internals/parsing/clexer.mll"
    ( if is_ghost_code () || is_oneline_ghost () then begin
        GHOST (currentLoc())
      end else begin
        let start = Lexing.lexeme_start_p lexbuf in
        let source = Cil_datatype.Position.of_lexing_pos start in
        E.parse_error ~source "Use of \\ghost out of ghost code"
      end
    )
# 2109 "src/kernel_internals/parsing/clexer.ml"

  | 5 ->
# 644 "src/kernel_internals/parsing/clexer.mll"
           (initial lexbuf)
# 2114 "src/kernel_internals/parsing/clexer.ml"

  | 6 ->
# 645 "src/kernel_internals/parsing/clexer.mll"
                                        ( E.newline ();
                                          if !pragmaLine then
                                            begin
                                              pragmaLine := false;
                                              PRAGMA_EOL
                                            end
                                          else if is_oneline_ghost () then
                                            begin
                                              exit_oneline_ghost ();
                                              RGHOST
                                            end
                                          else begin
                                            initial lexbuf
                                          end )
# 2132 "src/kernel_internals/parsing/clexer.ml"

  | 7 ->
# 659 "src/kernel_internals/parsing/clexer.mll"
                                        ( E.newline ();
                                          initial lexbuf
                                        )
# 2139 "src/kernel_internals/parsing/clexer.ml"

  | 8 ->
# 662 "src/kernel_internals/parsing/clexer.mll"
         ( hash lexbuf)
# 2144 "src/kernel_internals/parsing/clexer.ml"

  | 9 ->
# 663 "src/kernel_internals/parsing/clexer.mll"
          ( hash lexbuf)
# 2149 "src/kernel_internals/parsing/clexer.ml"

  | 10 ->
# 664 "src/kernel_internals/parsing/clexer.mll"
                                   ( PRAGMA (currentLoc ()) )
# 2154 "src/kernel_internals/parsing/clexer.ml"

  | 11 ->
# 665 "src/kernel_internals/parsing/clexer.mll"
          (
      let start = Lexing.lexeme_start_p lexbuf in
      let content = chr lexbuf in
      let last = Lexing.lexeme_end_p lexbuf in
      CST_CHAR (content, Cil_datatype.Location.of_lexing_loc (start,last))
    )
# 2164 "src/kernel_internals/parsing/clexer.ml"

  | 12 ->
# 671 "src/kernel_internals/parsing/clexer.mll"
          (
      let start = Lexing.lexeme_start_p lexbuf in
      let content = chr lexbuf in
      let last = Lexing.lexeme_end_p lexbuf in
      CST_WCHAR (content, Cil_datatype.Location.of_lexing_loc (start,last))
    )
# 2174 "src/kernel_internals/parsing/clexer.ml"

  | 13 ->
# 677 "src/kernel_internals/parsing/clexer.mll"
         (
      let start = Lexing.lexeme_start_p lexbuf in
      let content = str lexbuf in
      let last = Lexing.lexeme_end_p lexbuf in
      CST_STRING (content, Cil_datatype.Location.of_lexing_loc (start,last))
    )
# 2184 "src/kernel_internals/parsing/clexer.ml"

  | 14 ->
# 683 "src/kernel_internals/parsing/clexer.mll"
           (
      let start = Lexing.lexeme_start_p lexbuf in
      let content = str lexbuf in
      let last = Lexing.lexeme_end_p lexbuf in
      CST_WSTRING(content, Cil_datatype.Location.of_lexing_loc (start,last))
    )
# 2194 "src/kernel_internals/parsing/clexer.ml"

  | 15 ->
# 689 "src/kernel_internals/parsing/clexer.mll"
             (CST_FLOAT (Lexing.lexeme lexbuf, currentLoc ()))
# 2199 "src/kernel_internals/parsing/clexer.ml"

  | 16 ->
# 690 "src/kernel_internals/parsing/clexer.mll"
                           ( (* GCC Extension for binary numbers *)
                                          CST_INT (Lexing.lexeme lexbuf, currentLoc ()))
# 2205 "src/kernel_internals/parsing/clexer.ml"

  | 17 ->
# 692 "src/kernel_internals/parsing/clexer.mll"
            (CST_INT (Lexing.lexeme lexbuf, currentLoc ()))
# 2210 "src/kernel_internals/parsing/clexer.ml"

  | 18 ->
# 693 "src/kernel_internals/parsing/clexer.mll"
            (CST_INT (Lexing.lexeme lexbuf, currentLoc ()))
# 2215 "src/kernel_internals/parsing/clexer.ml"

  | 19 ->
# 694 "src/kernel_internals/parsing/clexer.mll"
            (CST_INT (Lexing.lexeme lexbuf, currentLoc ()))
# 2220 "src/kernel_internals/parsing/clexer.ml"

  | 20 ->
# 695 "src/kernel_internals/parsing/clexer.mll"
             (EOF)
# 2225 "src/kernel_internals/parsing/clexer.ml"

  | 21 ->
# 696 "src/kernel_internals/parsing/clexer.mll"
           (ELLIPSIS)
# 2230 "src/kernel_internals/parsing/clexer.ml"

  | 22 ->
# 697 "src/kernel_internals/parsing/clexer.mll"
          (PLUS_EQ)
# 2235 "src/kernel_internals/parsing/clexer.ml"

  | 23 ->
# 698 "src/kernel_internals/parsing/clexer.mll"
          (MINUS_EQ)
# 2240 "src/kernel_internals/parsing/clexer.ml"

  | 24 ->
# 699 "src/kernel_internals/parsing/clexer.mll"
          (STAR_EQ)
# 2245 "src/kernel_internals/parsing/clexer.ml"

  | 25 ->
# 700 "src/kernel_internals/parsing/clexer.mll"
          (SLASH_EQ)
# 2250 "src/kernel_internals/parsing/clexer.ml"

  | 26 ->
# 701 "src/kernel_internals/parsing/clexer.mll"
          (PERCENT_EQ)
# 2255 "src/kernel_internals/parsing/clexer.ml"

  | 27 ->
# 702 "src/kernel_internals/parsing/clexer.mll"
          (PIPE_EQ)
# 2260 "src/kernel_internals/parsing/clexer.ml"

  | 28 ->
# 703 "src/kernel_internals/parsing/clexer.mll"
          (AND_EQ)
# 2265 "src/kernel_internals/parsing/clexer.ml"

  | 29 ->
# 704 "src/kernel_internals/parsing/clexer.mll"
          (CIRC_EQ)
# 2270 "src/kernel_internals/parsing/clexer.ml"

  | 30 ->
# 705 "src/kernel_internals/parsing/clexer.mll"
           (INF_INF_EQ)
# 2275 "src/kernel_internals/parsing/clexer.ml"

  | 31 ->
# 706 "src/kernel_internals/parsing/clexer.mll"
           (SUP_SUP_EQ)
# 2280 "src/kernel_internals/parsing/clexer.ml"

  | 32 ->
# 707 "src/kernel_internals/parsing/clexer.mll"
          (INF_INF)
# 2285 "src/kernel_internals/parsing/clexer.ml"

  | 33 ->
# 708 "src/kernel_internals/parsing/clexer.mll"
          (SUP_SUP)
# 2290 "src/kernel_internals/parsing/clexer.ml"

  | 34 ->
# 709 "src/kernel_internals/parsing/clexer.mll"
           (EQ_EQ)
# 2295 "src/kernel_internals/parsing/clexer.ml"

  | 35 ->
# 710 "src/kernel_internals/parsing/clexer.mll"
           (EXCLAM_EQ)
# 2300 "src/kernel_internals/parsing/clexer.ml"

  | 36 ->
# 711 "src/kernel_internals/parsing/clexer.mll"
          (INF_EQ)
# 2305 "src/kernel_internals/parsing/clexer.ml"

  | 37 ->
# 712 "src/kernel_internals/parsing/clexer.mll"
          (SUP_EQ)
# 2310 "src/kernel_internals/parsing/clexer.ml"

  | 38 ->
# 713 "src/kernel_internals/parsing/clexer.mll"
         (EQ)
# 2315 "src/kernel_internals/parsing/clexer.ml"

  | 39 ->
# 714 "src/kernel_internals/parsing/clexer.mll"
         (INF)
# 2320 "src/kernel_internals/parsing/clexer.ml"

  | 40 ->
# 715 "src/kernel_internals/parsing/clexer.mll"
         (SUP)
# 2325 "src/kernel_internals/parsing/clexer.ml"

  | 41 ->
# 716 "src/kernel_internals/parsing/clexer.mll"
          (PLUS_PLUS (currentLoc ()))
# 2330 "src/kernel_internals/parsing/clexer.ml"

  | 42 ->
# 717 "src/kernel_internals/parsing/clexer.mll"
          (MINUS_MINUS (currentLoc ()))
# 2335 "src/kernel_internals/parsing/clexer.ml"

  | 43 ->
# 718 "src/kernel_internals/parsing/clexer.mll"
          (ARROW)
# 2340 "src/kernel_internals/parsing/clexer.ml"

  | 44 ->
# 719 "src/kernel_internals/parsing/clexer.mll"
         (PLUS (currentLoc ()))
# 2345 "src/kernel_internals/parsing/clexer.ml"

  | 45 ->
# 720 "src/kernel_internals/parsing/clexer.mll"
         (MINUS (currentLoc ()))
# 2350 "src/kernel_internals/parsing/clexer.ml"

  | 46 ->
# 722 "src/kernel_internals/parsing/clexer.mll"
                    ( if is_ghost_code () then might_end_ghost lexbuf
                      else
                        STAR (currentLoc ()))
# 2357 "src/kernel_internals/parsing/clexer.ml"

  | 47 ->
let
# 725 "src/kernel_internals/parsing/clexer.mll"
                   c
# 2363 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 1) in
# 726 "src/kernel_internals/parsing/clexer.mll"
   ( if c = !annot_char then
       if is_ghost_code () || is_oneline_ghost () then begin
         enter_ghost_annot();
         annot_lex initial annot_first_token lexbuf
       end else
         E.parse_error "This kind of annotation is valid only inside ghost code"
     else begin
       lexbuf.Lexing.lex_curr_pos <- lexbuf.Lexing.lex_curr_pos - 1;
       SLASH
     end )
# 2376 "src/kernel_internals/parsing/clexer.ml"

  | 48 ->
# 736 "src/kernel_internals/parsing/clexer.mll"
                                        (SLASH)
# 2381 "src/kernel_internals/parsing/clexer.ml"

  | 49 ->
# 737 "src/kernel_internals/parsing/clexer.mll"
         (PERCENT)
# 2386 "src/kernel_internals/parsing/clexer.ml"

  | 50 ->
# 738 "src/kernel_internals/parsing/clexer.mll"
         (EXCLAM (currentLoc ()))
# 2391 "src/kernel_internals/parsing/clexer.ml"

  | 51 ->
# 739 "src/kernel_internals/parsing/clexer.mll"
          (AND_AND (currentLoc ()))
# 2396 "src/kernel_internals/parsing/clexer.ml"

  | 52 ->
# 740 "src/kernel_internals/parsing/clexer.mll"
          (PIPE_PIPE)
# 2401 "src/kernel_internals/parsing/clexer.ml"

  | 53 ->
# 741 "src/kernel_internals/parsing/clexer.mll"
         (AND (currentLoc ()))
# 2406 "src/kernel_internals/parsing/clexer.ml"

  | 54 ->
# 742 "src/kernel_internals/parsing/clexer.mll"
         (PIPE)
# 2411 "src/kernel_internals/parsing/clexer.ml"

  | 55 ->
# 743 "src/kernel_internals/parsing/clexer.mll"
         (CIRC)
# 2416 "src/kernel_internals/parsing/clexer.ml"

  | 56 ->
# 744 "src/kernel_internals/parsing/clexer.mll"
         (QUEST)
# 2421 "src/kernel_internals/parsing/clexer.ml"

  | 57 ->
# 746 "src/kernel_internals/parsing/clexer.mll"
  ( if Cabshelper.is_attr_test () then begin
        Cabshelper.pop_attr_test (); COLON2
      end else COLON
  )
# 2429 "src/kernel_internals/parsing/clexer.ml"

  | 58 ->
# 750 "src/kernel_internals/parsing/clexer.mll"
                (TILDE (currentLoc ()))
# 2434 "src/kernel_internals/parsing/clexer.ml"

  | 59 ->
# 752 "src/kernel_internals/parsing/clexer.mll"
               (dbgToken (LBRACE (currentLoc ())))
# 2439 "src/kernel_internals/parsing/clexer.ml"

  | 60 ->
# 753 "src/kernel_internals/parsing/clexer.mll"
               (dbgToken (RBRACE (currentLoc ())))
# 2444 "src/kernel_internals/parsing/clexer.ml"

  | 61 ->
# 754 "src/kernel_internals/parsing/clexer.mll"
                (dbgToken (LBRACE (currentLoc ())))
# 2449 "src/kernel_internals/parsing/clexer.ml"

  | 62 ->
# 755 "src/kernel_internals/parsing/clexer.mll"
                (dbgToken (RBRACE (currentLoc ())))
# 2454 "src/kernel_internals/parsing/clexer.ml"

  | 63 ->
# 756 "src/kernel_internals/parsing/clexer.mll"
          (LBRACKET)
# 2459 "src/kernel_internals/parsing/clexer.ml"

  | 64 ->
# 757 "src/kernel_internals/parsing/clexer.mll"
          (RBRACKET)
# 2464 "src/kernel_internals/parsing/clexer.ml"

  | 65 ->
# 758 "src/kernel_internals/parsing/clexer.mll"
           (LBRACKET)
# 2469 "src/kernel_internals/parsing/clexer.ml"

  | 66 ->
# 759 "src/kernel_internals/parsing/clexer.mll"
           (RBRACKET)
# 2474 "src/kernel_internals/parsing/clexer.ml"

  | 67 ->
# 760 "src/kernel_internals/parsing/clexer.mll"
               (dbgToken (LPAREN (currentLoc ())) )
# 2479 "src/kernel_internals/parsing/clexer.ml"

  | 68 ->
# 761 "src/kernel_internals/parsing/clexer.mll"
          (RPAREN)
# 2484 "src/kernel_internals/parsing/clexer.ml"

  | 69 ->
# 762 "src/kernel_internals/parsing/clexer.mll"
               (dbgToken (SEMICOLON (currentLoc ())) )
# 2489 "src/kernel_internals/parsing/clexer.ml"

  | 70 ->
# 763 "src/kernel_internals/parsing/clexer.mll"
          (COMMA)
# 2494 "src/kernel_internals/parsing/clexer.ml"

  | 71 ->
# 764 "src/kernel_internals/parsing/clexer.mll"
          (DOT)
# 2499 "src/kernel_internals/parsing/clexer.ml"

  | 72 ->
# 765 "src/kernel_internals/parsing/clexer.mll"
             (SIZEOF (currentLoc ()))
# 2504 "src/kernel_internals/parsing/clexer.ml"

  | 73 ->
# 766 "src/kernel_internals/parsing/clexer.mll"
                                        ( if !Cprint.msvcMode then
                                             MSASM (msasm lexbuf, currentLoc ())
                                          else (ASM (currentLoc ())) )
# 2511 "src/kernel_internals/parsing/clexer.ml"

  | 74 ->
# 771 "src/kernel_internals/parsing/clexer.mll"
                                        ( let _ = matchingpars 0 lexbuf in
                                          initial lexbuf
                                        )
# 2518 "src/kernel_internals/parsing/clexer.ml"

  | 75 ->
# 777 "src/kernel_internals/parsing/clexer.mll"
                                        (initial lexbuf )
# 2523 "src/kernel_internals/parsing/clexer.ml"

  | 76 ->
# 778 "src/kernel_internals/parsing/clexer.mll"
                                        (scan_ident (Lexing.lexeme lexbuf))
# 2528 "src/kernel_internals/parsing/clexer.ml"

  | 77 ->
# 780 "src/kernel_internals/parsing/clexer.mll"
  ( if is_oneline_ghost() then begin
      exit_oneline_ghost (); RGHOST
    end
    else EOF
  )
# 2537 "src/kernel_internals/parsing/clexer.ml"

  | 78 ->
let
# 785 "src/kernel_internals/parsing/clexer.mll"
                     c
# 2543 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 786 "src/kernel_internals/parsing/clexer.mll"
  ( if is_ghost_code() && c = '@' then initial lexbuf
    else E.parse_error "Invalid symbol"
  )
# 2549 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_initial_rec lexbuf __ocaml_lex_state

and might_end_ghost lexbuf =
   __ocaml_lex_might_end_ghost_rec lexbuf 183
and __ocaml_lex_might_end_ghost_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 790 "src/kernel_internals/parsing/clexer.mll"
        ( exit_ghost_code(); RGHOST )
# 2561 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 791 "src/kernel_internals/parsing/clexer.mll"
       ( STAR (currentLoc()) )
# 2566 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_might_end_ghost_rec lexbuf __ocaml_lex_state

and comment buffer lexbuf =
   __ocaml_lex_comment_rec buffer lexbuf 185
and __ocaml_lex_comment_rec buffer lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 794 "src/kernel_internals/parsing/clexer.mll"
                (  )
# 2578 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 795 "src/kernel_internals/parsing/clexer.mll"
                ( E.parse_error "Unterminated C comment" )
# 2583 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 796 "src/kernel_internals/parsing/clexer.mll"
                ( lex_comment comment buffer lexbuf )
# 2588 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_comment_rec buffer lexbuf __ocaml_lex_state

and onelinecomment buffer lexbuf =
   __ocaml_lex_onelinecomment_rec buffer lexbuf 190
and __ocaml_lex_onelinecomment_rec buffer lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 799 "src/kernel_internals/parsing/clexer.mll"
                ( if is_ghost_code () then
                      (* end of multiline comment *)
                      lexbuf.Lexing.lex_curr_pos <-
                          lexbuf.Lexing.lex_curr_pos - 2
                  else
                      lex_comment onelinecomment buffer lexbuf
                )
# 2606 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 806 "src/kernel_internals/parsing/clexer.mll"
                (  )
# 2611 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 807 "src/kernel_internals/parsing/clexer.mll"
                ( lex_comment onelinecomment buffer lexbuf )
# 2616 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_onelinecomment_rec buffer lexbuf __ocaml_lex_state

and matchingpars parsopen lexbuf =
   __ocaml_lex_matchingpars_rec parsopen lexbuf 195
and __ocaml_lex_matchingpars_rec parsopen lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 810 "src/kernel_internals/parsing/clexer.mll"
                ( E.newline (); matchingpars parsopen lexbuf )
# 2628 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 811 "src/kernel_internals/parsing/clexer.mll"
                ( matchingpars parsopen lexbuf )
# 2633 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 812 "src/kernel_internals/parsing/clexer.mll"
                ( matchingpars (parsopen + 1) lexbuf )
# 2638 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
# 813 "src/kernel_internals/parsing/clexer.mll"
                ( if parsopen > 1 then
                    matchingpars (parsopen - 1) lexbuf
                )
# 2645 "src/kernel_internals/parsing/clexer.ml"

  | 4 ->
# 816 "src/kernel_internals/parsing/clexer.mll"
         ( do_lex_comment comment lexbuf ;
                  matchingpars parsopen lexbuf )
# 2651 "src/kernel_internals/parsing/clexer.ml"

  | 5 ->
# 818 "src/kernel_internals/parsing/clexer.mll"
        ( let _ = str lexbuf in
                  matchingpars parsopen lexbuf )
# 2657 "src/kernel_internals/parsing/clexer.ml"

  | 6 ->
# 820 "src/kernel_internals/parsing/clexer.mll"
                ( matchingpars parsopen lexbuf )
# 2662 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_matchingpars_rec parsopen lexbuf __ocaml_lex_state

and hash lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 2 (-1); __ocaml_lex_hash_rec lexbuf 204
and __ocaml_lex_hash_rec lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 824 "src/kernel_internals/parsing/clexer.mll"
        ( E.newline (); initial lexbuf)
# 2674 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 825 "src/kernel_internals/parsing/clexer.mll"
         ( hash lexbuf)
# 2679 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 826 "src/kernel_internals/parsing/clexer.mll"
         ( (* We are seeing a line number. This is the number for the
                   * next line *)
                 let s = Lexing.lexeme lexbuf in
                 let lineno = try
                   int_of_string s
                 with Failure _ ->
                   (* the int is too big. *)
                   Kernel.warning "Bad line number in preprocessed file: %s" s;
                   (-1)
                 in
                 E.setCurrentLine (lineno - 1);
                  (* A file name may follow *)
		  file lexbuf )
# 2696 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
# 839 "src/kernel_internals/parsing/clexer.mll"
                ( hash lexbuf )
# 2701 "src/kernel_internals/parsing/clexer.ml"

  | 4 ->
let
# 842 "src/kernel_internals/parsing/clexer.mll"
                                     pragmaName
# 2707 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 843 "src/kernel_internals/parsing/clexer.mll"
                ( let here = currentLoc () in
                  PRAGMA_LINE (pragmaName ^ pragma lexbuf, here)
                )
# 2713 "src/kernel_internals/parsing/clexer.ml"

  | 5 ->
# 846 "src/kernel_internals/parsing/clexer.mll"
                ( pragmaLine := true; PRAGMA (currentLoc ()) )
# 2718 "src/kernel_internals/parsing/clexer.ml"

  | 6 ->
# 847 "src/kernel_internals/parsing/clexer.mll"
            ( endline lexbuf)
# 2723 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_hash_rec lexbuf __ocaml_lex_state

and file lexbuf =
   __ocaml_lex_file_rec lexbuf 311
and __ocaml_lex_file_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 850 "src/kernel_internals/parsing/clexer.mll"
                      (E.newline (); initial lexbuf)
# 2735 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 851 "src/kernel_internals/parsing/clexer.mll"
          (file lexbuf)
# 2740 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
let
# 854 "src/kernel_internals/parsing/clexer.mll"
                                     d
# 2746 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 1) (lexbuf.Lexing.lex_curr_pos + -3) in
# 854 "src/kernel_internals/parsing/clexer.mll"
                                               (
        E.setCurrentWorkingDirectory d;
                                 endline lexbuf )
# 2752 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
let
# 857 "src/kernel_internals/parsing/clexer.mll"
                                        f
# 2758 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 1) (lexbuf.Lexing.lex_curr_pos + -1) in
# 857 "src/kernel_internals/parsing/clexer.mll"
                                               (
        let unescape = Str.regexp_string "\\\"" in
        let f = Str.global_replace unescape "\"" f in
                                 E.setCurrentFile f;
				 endline lexbuf)
# 2766 "src/kernel_internals/parsing/clexer.ml"

  | 4 ->
# 863 "src/kernel_internals/parsing/clexer.mll"
      (endline lexbuf)
# 2771 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_file_rec lexbuf __ocaml_lex_state

and endline lexbuf =
   __ocaml_lex_endline_rec lexbuf 325
and __ocaml_lex_endline_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 866 "src/kernel_internals/parsing/clexer.mll"
                ( E.newline (); initial lexbuf)
# 2783 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 867 "src/kernel_internals/parsing/clexer.mll"
                                ( EOF )
# 2788 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 868 "src/kernel_internals/parsing/clexer.mll"
      ( endline lexbuf)
# 2793 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_endline_rec lexbuf __ocaml_lex_state

and pragma lexbuf =
   __ocaml_lex_pragma_rec lexbuf 329
and __ocaml_lex_pragma_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 871 "src/kernel_internals/parsing/clexer.mll"
                        ( E.newline (); "" )
# 2805 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 872 "src/kernel_internals/parsing/clexer.mll"
                        ( let cur = Lexing.lexeme lexbuf in
                          cur ^ (pragma lexbuf) )
# 2811 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_pragma_rec lexbuf __ocaml_lex_state

and str lexbuf =
   __ocaml_lex_str_rec lexbuf 332
and __ocaml_lex_str_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 876 "src/kernel_internals/parsing/clexer.mll"
                        ([])
# 2823 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 877 "src/kernel_internals/parsing/clexer.mll"
             (lex_hex_escape str lexbuf)
# 2828 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 878 "src/kernel_internals/parsing/clexer.mll"
             (lex_oct_escape str lexbuf)
# 2833 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
# 879 "src/kernel_internals/parsing/clexer.mll"
          (lex_simple_escape str lexbuf)
# 2838 "src/kernel_internals/parsing/clexer.ml"

  | 4 ->
# 880 "src/kernel_internals/parsing/clexer.mll"
                        (E.parse_error "unterminated string" )
# 2843 "src/kernel_internals/parsing/clexer.ml"

  | 5 ->
# 881 "src/kernel_internals/parsing/clexer.mll"
     (lex_unescaped str lexbuf)
# 2848 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_str_rec lexbuf __ocaml_lex_state

and chr lexbuf =
   __ocaml_lex_chr_rec lexbuf 343
and __ocaml_lex_chr_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 884 "src/kernel_internals/parsing/clexer.mll"
              ([])
# 2860 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 885 "src/kernel_internals/parsing/clexer.mll"
             (lex_hex_escape chr lexbuf)
# 2865 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 886 "src/kernel_internals/parsing/clexer.mll"
             (lex_oct_escape chr lexbuf)
# 2870 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
# 887 "src/kernel_internals/parsing/clexer.mll"
          (lex_simple_escape chr lexbuf)
# 2875 "src/kernel_internals/parsing/clexer.ml"

  | 4 ->
# 888 "src/kernel_internals/parsing/clexer.mll"
                        ( E.parse_error "unterminated char" )
# 2880 "src/kernel_internals/parsing/clexer.ml"

  | 5 ->
# 889 "src/kernel_internals/parsing/clexer.mll"
     (lex_unescaped chr lexbuf)
# 2885 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_chr_rec lexbuf __ocaml_lex_state

and msasm lexbuf =
   __ocaml_lex_msasm_rec lexbuf 354
and __ocaml_lex_msasm_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 892 "src/kernel_internals/parsing/clexer.mll"
                        ( msasm lexbuf )
# 2897 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 893 "src/kernel_internals/parsing/clexer.mll"
                        ( msasminbrace lexbuf )
# 2902 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 894 "src/kernel_internals/parsing/clexer.mll"
                        ( let cur = Lexing.lexeme lexbuf in
                          cur ^ (msasmnobrace lexbuf) )
# 2908 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_msasm_rec lexbuf __ocaml_lex_state

and msasminbrace lexbuf =
   __ocaml_lex_msasminbrace_rec lexbuf 358
and __ocaml_lex_msasminbrace_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 898 "src/kernel_internals/parsing/clexer.mll"
                        ( "" )
# 2920 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 899 "src/kernel_internals/parsing/clexer.mll"
                        ( let cur = Lexing.lexeme lexbuf in
                          cur ^ (msasminbrace lexbuf) )
# 2926 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_msasminbrace_rec lexbuf __ocaml_lex_state

and msasmnobrace lexbuf =
   __ocaml_lex_msasmnobrace_rec lexbuf 361
and __ocaml_lex_msasmnobrace_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 902 "src/kernel_internals/parsing/clexer.mll"
                        ( lexbuf.Lexing.lex_curr_pos <-
                               lexbuf.Lexing.lex_curr_pos - 1;
                          "" )
# 2940 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 905 "src/kernel_internals/parsing/clexer.mll"
                        ( lexbuf.Lexing.lex_curr_pos <-
                               lexbuf.Lexing.lex_curr_pos - 5;
                          "" )
# 2947 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 908 "src/kernel_internals/parsing/clexer.mll"
                        ( let cur = Lexing.lexeme lexbuf in

                          cur ^ (msasmnobrace lexbuf) )
# 2954 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_msasmnobrace_rec lexbuf __ocaml_lex_state

and annot_first_token lexbuf =
   __ocaml_lex_annot_first_token_rec lexbuf 369
and __ocaml_lex_annot_first_token_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 913 "src/kernel_internals/parsing/clexer.mll"
                                                     comments
# 2967 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 5) (lexbuf.Lexing.lex_curr_pos + -4) in
# 913 "src/kernel_internals/parsing/clexer.mll"
                                                                      (
      if is_oneline_ghost () then E.parse_error "nested ghost code";
      Buffer.clear buf;
      let loc = currentLoc () in
      do_ghost_else_comments true comments ;
      enter_ghost_code ();
      LGHOST_ELSE (loc)
    )
# 2978 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 921 "src/kernel_internals/parsing/clexer.mll"
            (
      if is_oneline_ghost () then E.parse_error "nested ghost code";
      Buffer.clear buf;
      enter_ghost_code ();
      LGHOST
    )
# 2988 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
let
# 927 "src/kernel_internals/parsing/clexer.mll"
                         c
# 2994 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 927 "src/kernel_internals/parsing/clexer.mll"
                           ( Buffer.add_char buf c; annot_first_token lexbuf )
# 2998 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
# 928 "src/kernel_internals/parsing/clexer.mll"
         ( E.newline(); Buffer.add_char buf '\n'; annot_first_token lexbuf )
# 3003 "src/kernel_internals/parsing/clexer.ml"

  | 4 ->
# 929 "src/kernel_internals/parsing/clexer.mll"
       ( annot_token lexbuf )
# 3008 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_annot_first_token_rec lexbuf __ocaml_lex_state

and annot_token lexbuf =
   __ocaml_lex_annot_token_rec lexbuf 395
and __ocaml_lex_annot_token_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 931 "src/kernel_internals/parsing/clexer.mll"
         ( if is_ghost_annot () then
             E.parse_error "Ghost multi-line annotation not terminated";
           let s = Buffer.contents buf in
           make_annot ~one_line:false initial lexbuf s )
# 3023 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 935 "src/kernel_internals/parsing/clexer.mll"
         ( E.parse_error "Unterminated annotation" )
# 3028 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 936 "src/kernel_internals/parsing/clexer.mll"
         (E.newline(); Buffer.add_char buf '\n'; annot_token lexbuf )
# 3033 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
let
# 937 "src/kernel_internals/parsing/clexer.mll"
         c
# 3039 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 937 "src/kernel_internals/parsing/clexer.mll"
           ( if is_ghost_annot () && c = !annot_char then
               might_end_ghost_annot lexbuf
             else (Buffer.add_char buf c; annot_token lexbuf) )
# 3045 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_annot_token_rec lexbuf __ocaml_lex_state

and might_end_ghost_annot lexbuf =
   __ocaml_lex_might_end_ghost_annot_rec lexbuf 401
and __ocaml_lex_might_end_ghost_annot_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 941 "src/kernel_internals/parsing/clexer.mll"
        ( exit_ghost_annot ();
          let s = Buffer.contents buf in
          make_annot ~one_line:false initial lexbuf s )
# 3059 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 944 "src/kernel_internals/parsing/clexer.mll"
       ( Buffer.add_char buf !annot_char; annot_token lexbuf )
# 3064 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_might_end_ghost_annot_rec lexbuf __ocaml_lex_state

and annot_one_line lexbuf =
   __ocaml_lex_annot_one_line_rec lexbuf 403
and __ocaml_lex_annot_one_line_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 946 "src/kernel_internals/parsing/clexer.mll"
                                comments
# 3077 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme lexbuf (lexbuf.Lexing.lex_start_pos + 5) (lexbuf.Lexing.lex_curr_pos + -4) in
# 946 "src/kernel_internals/parsing/clexer.mll"
                                                 (
      do_ghost_else_comments false comments ;
      if is_oneline_ghost () then E.parse_error "nested ghost code";
      enter_oneline_ghost (); LGHOST_ELSE (currentLoc ())
  )
# 3085 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 951 "src/kernel_internals/parsing/clexer.mll"
            (
      if is_oneline_ghost () then E.parse_error "nested ghost code";
      enter_oneline_ghost (); LGHOST
    )
# 3093 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
let
# 955 "src/kernel_internals/parsing/clexer.mll"
                         c
# 3099 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 955 "src/kernel_internals/parsing/clexer.mll"
                           ( Buffer.add_char buf c; annot_one_line lexbuf )
# 3103 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
# 956 "src/kernel_internals/parsing/clexer.mll"
       ( annot_one_line_logic lexbuf )
# 3108 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_annot_one_line_rec lexbuf __ocaml_lex_state

and annot_one_line_logic lexbuf =
   __ocaml_lex_annot_one_line_logic_rec lexbuf 416
and __ocaml_lex_annot_one_line_logic_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 958 "src/kernel_internals/parsing/clexer.mll"
         ( make_annot ~one_line:true initial lexbuf (Buffer.contents buf) )
# 3120 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 959 "src/kernel_internals/parsing/clexer.mll"
        ( E.parse_error "Invalid C file: should end with a newline" )
# 3125 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
let
# 960 "src/kernel_internals/parsing/clexer.mll"
         c
# 3131 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 960 "src/kernel_internals/parsing/clexer.mll"
           ( Buffer.add_char buf c; annot_one_line_logic lexbuf )
# 3135 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_annot_one_line_logic_rec lexbuf __ocaml_lex_state

;;

# 962 "src/kernel_internals/parsing/clexer.mll"
 

  (* Catch the exceptions raised by the lexer itself *)
  let initial lexbuf =
    try initial lexbuf
    with Failure _ -> raise Parsing.Parse_error


# 3151 "src/kernel_internals/parsing/clexer.ml"
